
import datetime as dt
import os
import shutil
import string
import sys

import numpy as np
import pandas as pd
import pytest
try:
    from pandas.tslib import Timestamp
except ImportError:
    from pandas import Timestamp
from six import PY2

from fastparquet import write, ParquetFile
from fastparquet.test.util import tempdir


def frame_symbol_dtTrade_type_strike(days=1 * 252,
                       start_date=dt.datetime(2005, 1, 1, hour=0, minute=0, second=0),
                       symbols=['SPY', 'FB', 'TLT'],
                       numbercolumns=1):
    base = start_date
    date_list = [base + dt.timedelta(days=x) for x in range(0, days)]
    tuple_list = []
    for x in symbols:
        for y in date_list:
            tuple_list.append((x, y.year, y))
    index = pd.MultiIndex.from_tuples(tuple_list, names=('symbol', 'year', 'dtTrade'))
    np.random.seed(seed=0)
    df = pd.DataFrame(np.random.randn(index.size, numbercolumns),
                      index=index, columns=[x for x in string.ascii_uppercase[0:numbercolumns]])
    return df

@pytest.mark.parametrize('input_symbols,input_days,file_scheme,input_columns,'
                         'partitions,filters',
                         [
                             (['NOW', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['now', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['TODAY', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['VIX*', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['QQQ*', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['QQQ!', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['Q%QQ', 'SPY', 'VIX'], 2 * 252, 'hive', 2,
                              ['symbol', 'year'], [('symbol', '==', 'SPY')]),
                             (['NOW', 'SPY', 'VIX'], 10, 'hive', 2,
                              ['symbol', 'dtTrade'], [('symbol', '==', 'SPY')]),
                             (['NOW', 'SPY', 'VIX'], 10, 'hive', 2,
                              ['symbol', 'dtTrade'],
                              [('dtTrade', '==',
                                '2005-01-02T00:00:00.000000000')]),
                             (['NOW', 'SPY', 'VIX'], 10, 'hive', 2,
                              ['symbol', 'dtTrade'],
                              [('dtTrade', '==',
                                Timestamp('2005-01-01 00:00:00'))]),
                         ]
                         )
def test_frame_write_read_verify(tempdir, input_symbols, input_days,
                                 file_scheme,
                                 input_columns, partitions, filters):
    if os.name == 'nt':
        pytest.xfail("Partitioning folder names contain special characters which are not supported on Windows")

    # Generate Temp Director for parquet Files
    fdir = str(tempdir)
    fname = os.path.join(fdir, 'test')

    # Generate Test Input Frame
    input_df = frame_symbol_dtTrade_type_strike(days=input_days,
                                                symbols=input_symbols,
                                                numbercolumns=input_columns)
    input_df.reset_index(inplace=True)
    write(fname, input_df, partition_on=partitions, file_scheme=file_scheme,
          compression='SNAPPY')

    # Read Back Whole Parquet Structure
    output_df = ParquetFile(fname).to_pandas()
    for col in output_df.columns:
        assert col in input_df.columns.values
    assert len(input_df) == len(output_df)

    # Read with filters
    filtered_output_df = ParquetFile(fname).to_pandas(filters=filters)

    # Filter Input Frame to Match What Should Be Expected from parquet read
    # Handle either string or non-string inputs / works for timestamps
    filterStrings = []
    for name, operator, value in filters:
        if isinstance(value, str):
            value = "'{}'".format(value)
        else:
            value = value.__repr__()
        filterStrings.append("{} {} {}".format(name, operator, value))
    filters_expression = " and ".join(filterStrings)
    filtered_input_df = input_df.query(filters_expression)

    # Check to Ensure Columns Match
    for col in filtered_output_df.columns:
        assert col in filtered_input_df.columns.values
    # Check to Ensure Number of Rows Match
    assert len(filtered_input_df) == len(filtered_output_df)

    # Clean Up
    shutil.rmtree(fdir, ignore_errors=True)



import io
import sys
import types
from thrift.protocol.TCompactProtocol import TCompactProtocolAccelerated as TCompactProtocol
from thrift.protocol.TProtocol import TProtocolException

from .parquet_thrift.parquet import ttypes as parquet_thrift
from .util import ParquetException


def read_thrift(file_obj, ttype):
    """Read a thrift structure from the given fo."""
    from thrift.transport.TTransport import TFileObjectTransport, TBufferedTransport
    starting_pos = file_obj.tell()

    # set up the protocol chain
    ft = TFileObjectTransport(file_obj)
    bufsize = 2 ** 16
    # for accelerated reading ensure that we wrap this so that the CReadable transport can be used.
    bt = TBufferedTransport(ft, bufsize)
    pin = TCompactProtocol(bt)

    # read out type
    obj = ttype()
    obj.read(pin)

    # The read will actually overshoot due to the buffering that thrift does. Seek backwards to the correct spot,.
    buffer_pos = bt.cstringio_buf.tell()
    ending_pos = file_obj.tell()
    blocks = ((ending_pos - starting_pos) // bufsize) - 1
    if blocks < 0:
        blocks = 0
    file_obj.seek(starting_pos + blocks * bufsize + buffer_pos)
    return obj


def write_thrift(fobj, thrift):
    """Write binary compact representation of thiftpy structured object

    Parameters
    ----------
    fobj: open file-like object (binary mode)
    thrift: thriftpy object to write

    Returns
    -------
    Number of bytes written
    """
    t0 = fobj.tell()
    pout = TCompactProtocol(fobj)
    try:
        thrift.write(pout)
        fail = False
    except TProtocolException as e:
        typ, val, tb = sys.exc_info()
        frames = []
        while tb is not None:
            frames.append(tb)
            tb = tb.tb_next
        frame = [tb for tb in frames if 'write_struct' in str(tb.tb_frame.f_code)]
        variables = frame[0].tb_frame.f_locals
        obj = variables['obj']
        name = variables['fname']
        fail = True
    if fail:
        raise ParquetException('Thrift parameter validation failure %s'
                               ' when writing: %s-> Field: %s' % (
            val.args[0], obj, name
        ))
    return fobj.tell() - t0


def is_thrift_item(item):
    return hasattr(item, 'thrift_spec') and hasattr(item, 'read')


def thrift_print(structure, offset=0):
    """
    Handy recursive text ouput for thrift structures
    """
    if not is_thrift_item(structure):
        return str(structure)
    s = str(structure.__class__) + '\n'
    for key in dir(structure):
        if key.startswith('_') or key in ['thrift_spec', 'read', 'write',
                                          'default_spec', 'validate']:
            continue
        s = s + ' ' * offset + key + ': ' + thrift_print(getattr(structure, key)
                                                         , offset+2) + '\n'
    return s


def bind_method(cls, name, func):
    if sys.version_info.major == 2:
        setattr(cls, name, types.MethodType(func, None, cls))
    else:
        setattr(cls, name, func)


for clsname in dir(parquet_thrift):
    if clsname[0].isupper():
        cls = getattr(parquet_thrift, clsname)
        bind_method(cls, '__repr__', thrift_print)


def getstate_method(ob):
    b = io.BytesIO()
    write_thrift(b, ob)
    b.seek(0)
    return b.read()


def copy_method(self):
    cls = type(self)
    out = cls.__new__(cls)
    out.__dict__ = self.__dict__.copy()
    return out


def setstate_method(self, state):
    b = io.BytesIO(state)
    out = read_thrift(b, type(self))
    self.__dict__ = out.__dict__


for cls in [parquet_thrift.FileMetaData,
            parquet_thrift.RowGroup,
            parquet_thrift.ColumnChunk]:
    bind_method(cls, '__getstate__', getstate_method)
    bind_method(cls, '__setstate__', setstate_method)
    bind_method(cls, '__copy__', copy_method)
